; -*- Scheme -*-

		;;; balls

(define 9ball-cue-ball 8ball-cue-ball)
(define 9ball-9ball 9)
(define 9ball-faul-counter 0)

		;;; table

(define (9ball-table-lookup table-size-id)
  (case table-size-id
    ((0) table/pockets/8ft)
    ((1) table/pockets/9ft)))

(define (9ball-setup-diamond table)
  (let ((foot-spot (table-foot-spot table)))

    (define (*r x)
      (* x (+ pocket-ball-radius (mm 0.01))))

    (define (place-ball-in-diamond bid pos)
      (let ((pos	(ball-in-diamond pos 3)))
	(array-map! pos *r pos)
	(array-map! pos + pos foot-spot)
	(place-ball/rr bid pos)))
    
    (place-ball-in-diamond 1 0)
    (place-ball-in-diamond 9 4)

    (for-each place-ball-in-diamond
     (unsort '(2 3 4 5 6 7 8))
     '(1 2 3 5 6 7 8 ))))


	;;; Game state:
	  ; 0  table-struct
	  ; 1  sorted list of balls on table
	  ; 2  ball 9. sunken
	  ; 3  number of mantinel x ball colission for shot without foul
	  ; 4  ball in hand ?
	  ; 5  cue needs respot

(define (9ball-gs-new ts)
  (vector ts
	  '(1 2 3 4 5 6 7 8 9) 
	  #f 
	  4
	  #f #f))

(define (9ball-gs-copy x)
  (vector
   (vector-ref x 0)
   (list-copy (vector-ref x 1))
   (vector-ref x 2)
   (vector-ref x 3)
   (vector-ref x 4)
   (vector-ref x 5)))

(define (9ball-gs-display x)
  (Display "Table name:\t" (table-name (vector-ref x 0)))
  (Display "Balls on tale:\t" (9ball-gs-balls x))
  (Display "Ball on:\t" (9ball-gs-ball-on x))
  (Display "9. ball sunken:\t" (9ball-gs-9-sunken? x))
  (Display "Nof mantinel x ball collisions:\t" (9ball-gs-mantinel x))
  (Display "Ball in hand:\t" (9ball-gs-ball-in-hand? x))
  (Display "Cue ball needs respot:\t" (9ball-gs-cue-ball-needs-respot? x)))

(define (9ball-gs-balls gs)
  (vector-ref gs 1))
  
(define (9ball-gs-ball-on gs)
  (car (9ball-gs-balls gs)))

(define (9ball-gs-remove-ball gs ball-id)
  (vector-set! gs 1 (list-remove-member (9ball-gs-balls gs) ball-id)))

(define (9ball-gs-9-reset! gs)
  (vector-set! gs 2 #f))

(define (9ball-gs-9-sunken! gs)
  (vector-set! gs 2 #t))

(define (9ball-gs-9-sunken? gs)
  (vector-ref gs 2))

(define (9ball-gs-mantinel gs)
  (vector-ref gs 3))

(define (9ball-gs-mantinel-dec! gs x)
  (vector-set! gs 3 (- (vector-ref gs 3) x)))

(define (9ball-gs-mantinel-reset! gs)
  (vector-set! gs 3 1))

(define (9ball-gs-ball-in-hand? gs)
  (vector-ref gs 4))

(define (9ball-gs-ball-in-hand-set! gs p)
  (vector-set! gs 4 p))

(define (9ball-gs-cue-ball-needs-respot? gs)
  (vector-ref gs 5))

(define (9ball-gs-set-cue-ball-needs-respot! gs)
  (vector-set! gs 5 #t))

(define (9ball-gs-reset-cue-ball-needs-respot! gs)
  (vector-set! gs 5 #f))

(define (9ball-resume-fauls player-id state)
  (unless (eq? state 'faul)
    (game-reset-score player-id 0))
  (when (< 2 (get-player-points player-id 9ball-faul-counter))
     (message "3 consecutive fauls - game lose")
     (game-lose)))

(define (9ball-faul! state)
  (unless (eq? state 'faul)
    (game-add-point 9ball-faul-counter))
  (game-faul)
  'faul)

(define (9ball-set-events state x)
  (apply game-set-event-mask!
    (case state
     ((start)   
      '(cue-ball-collision ball-jump-off ball-pocketed ball-ball-collision 1))
     ((after-bb) 
      (cons 'ball-mantinel-collision 
	    (cons  x 
		   '(cue-ball-collision ball-pocketed ball-jump-off)))))))

(pool-rules "9-ball"

	    'number-of-balls		10
	    'number-of-players		2

	    'help-file			"game_9ball.html"


	    'game-state-deep-copy	9ball-gs-copy
	    'safe-to-undo		#t

	    'score		        '((faul-counter		"Fauls"))

	    'queries			'((player))

	    'options			'((group "Table size"
					   (one-of 1
						   "8ft table"
						   "9ft table")))
	    'table-setup
	      (lambda (table-size-id)
		(8ball-skins 10)
		(let ((table (9ball-table-lookup table-size-id)))
		  (table-new table 'foot-spot 'head-spot 'kitchen)
		  (cushion-height (+ pocket-ball-radius (cm 0.2)))))

	    'new-game
	      (lambda (table-size-id)
		(let ((table  (9ball-table-lookup table-size-id)))
		  
		  (game-state-set! (9ball-gs-new table))
		  (9ball-setup-diamond table)

		  (place-ball 0 (table-head-spot table))

		  (place-cue 0 #(-90 0))
		  (ball-can-move 9ball-cue-ball
				 (table-in-the-kitchen-pred table))))

	    'end-of-shot
	      (lambda (last-state game-state balls-out player-id)
		(DebugDisplayEndOfShot last-state game-state balls-out)
			;;; clear message which ball player is on
		(message-end)
			;;; disable all balls moving....
		(ball-can-move #f)

			;;; player have to pot ball or drive at
			  ; least some ball to mantinel
		(unless (or (eq? last-state 'ball-potted)
			    (>= 0 (9ball-gs-mantinel game-state)))
		  (9ball-faul! last-state)
		  (set! last-state 'faul))

			;;; count player fauls (3 successive -=> lose game)
		(9ball-resume-fauls player-id last-state)
		(9ball-gs-ball-in-hand-set! game-state (eq? last-state 'faul))

		(case last-state
		  ((faul)
			;;; respot nine ball
		   (if (9ball-gs-9-sunken? game-state)
		       (ball-respot 9ball-9ball 
				    (gs-table-foot-spot game-state)
				    0
				    (gs-table-height game-state)))
		   (if (9ball-gs-cue-ball-needs-respot? game-state)
		       (ball-respot-area 9ball-cue-ball
					 0 0
					 (gs-table-width game-state)
					 (gs-table-head game-state)))
		   (game-next-player))
		  ((ball-potted)
		   (and (9ball-gs-9-sunken? game-state)
			(game-win)))
		  (else
		   (game-next-player)))

		(9ball-gs-mantinel-reset! game-state)
		(9ball-gs-9-reset! game-state)
		(9ball-gs-reset-cue-ball-needs-respot! game-state))

	    'next-shot
	      (lambda (player-id game-state)
			;; inform player which ball he plays on
		(game-set-ball-on (9ball-gs-ball-on game-state))
		(message-start (string-append 
				 "You are on " 
				 (8ball-name (9ball-gs-ball-on game-state))))
			;; reset 9ball sunken state
		(9ball-gs-9-reset! game-state)
			;; reset events
		(9ball-set-events 'start 0)
			;; ball in hand - put cue ball anywhere you want
		(if (9ball-gs-ball-in-hand? game-state)
		    (ball-can-move 9ball-cue-ball)))

	    'event-handler
	      (lambda (event-type info1 info2 state game-state player-id)
		(DebugDisplayEvent event-type info1 info2 state game-state)
		(case event-type

		  ((cue-ball-collision)
		   (if (and (not state) (eq? info1 9ball-cue-ball))
		     'after-cue-ball-collision
		     (9ball-faul! state)))

		  ((ball-jump-off)
		   (cond
		    ((eq? info1 9ball-9ball)
			;; a little bit tricky - we say that it is sunken, but
			 ; because of faul it will be respotted...
		     (9ball-gs-9-sunken! game-state))
		    ((eq? info1 9ball-cue-ball)
			 ; white ball should be respotted on table (and than be moveable)
		     (9ball-gs-set-cue-ball-needs-respot game-state))
		    (else
			; remove ball from table....
		     (9ball-gs-remove-ball game-state info 1)))
		   (9ball-faul! state))

		  ((ball-pocketed)
		   (cond
		     ((eq? info1 9ball-cue-ball)
		      (9ball-gs-set-cue-ball-needs-respot! game-state)
		      (9ball-faul! state))
		     (else
		      (if (eq? info1 9ball-9ball)
			  (9ball-gs-9-sunken! game-state)
			  (9ball-gs-remove-ball game-state info1))
		      (case state
			((faul) 'faul)
			(else
			 (game-reward)
			 'ball-potted)))))

		  ((ball-ball-collision)
		   (if (eq? state 'faul)
		       'faul
		       (let ((bid (if (eq? info1 9ball-cue-ball)
				      info2
				      info1)))
			 (cond
			  ((eq? bid (9ball-gs-ball-on game-state))
			   (9ball-set-events 'after-bb (9ball-gs-mantinel game-state))
			   'after-bb)
			  (else
			   (9ball-faul! state))))))
		  ((ball-mantinel-collision)
		   (9ball-gs-mantinel-dec! game-state 1)))))

