; -*- Scheme -*-

(define carom-cue-ball-1 0)
(define carom-cue-ball-2 1)
(define carom-red-ball   2)

  
(define carom-ball-positions
   (vector
    'carom-spot-1
    'carom-spot-2
    'foot-spot))


(pool-rules "Carom billiard"

 	  'number-of-balls 3
	  'number-of-players 2

		;;; game-state of carom rules is very simple
		  ; so it is safe to undo call
	  'safe-to-undo	#t

	  'help-file "game_carom.html"

	  'score
	    '((counter		"Points"))

	  'options
	    '((group "Carom options"
	        (integer "Number of caroms to win a game" 20 1 999)))

	  'table-setup
 	    (lambda (max-carom)
	      (table-new table/carom 
			 'center-spot 'foot-spot 'head-spot
			 'carom-spot-1 'carom-spot-2)

	      (cushion-height (+ carom-ball-radius (cm .2)))

	      (ball-set-skin carom-cue-ball-1 "white ball" 'white-ball 'carom-ball)
	      (ball-set-skin carom-cue-ball-2 "ball with mark" 'white-ball-with-black-mark 'carom-ball)
	      (ball-set-skin carom-red-ball   "red ball" 'red-ball 'carom-ball))

	  'new-game
	    (lambda (max-carom)
	      (game-state-set! max-carom)

	      (place-ball carom-cue-ball-1 (table-spot table/carom 'carom-spot-1))
	      (place-ball carom-cue-ball-2 (table-head-spot table/carom))
	      (place-ball carom-red-ball   (table-foot-spot table/carom))

	      (place-cue 0 #(-90 0)))
			
	  'next-shot
	    (lambda (player-id game-state)
	      (game-set-ball-on player-id)
	      (place-cue player-id))

	  'end-of-shot
	    (lambda (last-state game-state balls-out-list player-id)
	      (DebugDisplayEndOfShot last-state game-state balls-out-list)
	      (cond 
	       ((eq? last-state 'carom)
		(if (>= (get-player-points) game-state)
		    (game-win)))

	       (else ; faul or not carom
		(game-next-player)))

		; process balls marked as out
	      (for-each
	        (lambda (ball-id)
		  (if (eq? ball-id carom-red-ball)
		      (or
		       (place-ball ball-id (table-foot-spot table/carom))
		       (place-ball ball-id (table-head-spot table/carom))
		       (place-ball ball-id (table-centerspot table/carom))
		      (or
		       (place-ball ball-id (table-head-spot table/carom))
		       (place-ball ball-id (table-foot-spot table/carom))
		       (place-ball ball-id (table-center-spot table/carom))))))
		(sort balls-out-list <)))

	  'queries '((player))
	  'event-mask '(ball-ball-collision cue-ball-collision ball-jump-off)

	  'event-handler
	    (lambda (event-type info1 info2 state game-state player-id)
	      (DebugDisplayEvent event-type info1 info2 state game-state "Pid" player-id)
	      (case event-type
		((ball-jump-off)
		 (game-faul)
		 (game-ball-out info1)
		 'faul)
		((cue-ball-collision)
		 (cond
		  ((or state				; start
		       (not (eq? player-id info1)))	; player 0 -=> shoot only to sphere 0
		   (game-faul)
		   'faul)
		  (else
		   (cons #f #f))))
		((ball-ball-collision)
		 (case state
		   ((faul) 'faul)
		   ((carom) 'carom)
		   (else ; player has not scored yet 
		    (let* ((s-player	(if (eq? player-id 0) 1 0))
			   (p-ball	(or (eq? info1 player-id) (eq? info2 player-id)))
			   (s-ball	(or (eq? info1 s-player) (eq? info2 s-player)))
			   (t-ball	(or (eq? info1 2) (eq? info2 2))))
		      (if p-ball
			  (cond
			   (s-ball (set-car! state #t))
			   (t-ball (set-cdr! state #t)))))
		    (cond 
		     ((and (car state) (cdr state))
		      (game-reward)
		      (game-add-point)
		      'carom)
		     (else
		      state))))))))
