; -*- Scheme -*-

; I have not played this game in real world - all rules and 
; proportions are taken from 
;
;    http://www.euronet.nl/users/cor/rules.html
;
; 0rfelyus

(define snooker-table-width (mm 1750))
(define snooker-table-height (mm 3500))

		; not so exact.... - I have to measure it somewhere ... :)

(define snooker-pocket-radius (cm 20)) ; snooker-ball-diameter)

(define snooker-Baulk (mm 700))
(define snooker-D-radius (mm 292))
(define snooker-D-radius^2 (* snooker-D-radius snooker-D-radius))

(define snooker-table-width/2 (/ snooker-table-width 2))
(define snooker-table-height/2 (/ snooker-table-height 2))
(define snooker-table-height*3/4 (* snooker-table-height 3/4))

(define (snooker-table-inside-D? x y)
  (let ((dx	(- snooker-table-width/2 x))
	 (dy	(- snooker-Baulk y)))
    (and (<= 0 dy) (>= snooker-D-radius^2
		       (+ (* dx dx) (* dy dy))))))
	;;; Spots

(define snooker-Spot 
  (vector snooker-table-width/2
	  (- snooker-table-height (mm 320))))

(define snooker-Center-spot
  (vector snooker-table-width/2
	  snooker-table-height/2))

(define snooker-Pyramid-spot
  (vector snooker-table-width/2
	  snooker-table-height*3/4))

(define snooker-Baulk-spot
  (vector snooker-table-width/2
	  snooker-Baulk))

(define snooker-green-ball-position
  (vector (- snooker-table-width/2 snooker-D-radius)
	  snooker-Baulk))

(define snooker-yellow-ball-position
  (vector (+ snooker-table-width/2 snooker-D-radius)
	  snooker-Baulk))

(define snooker-top-red-position
  (vector snooker-table-width/2
	  (+ snooker-table-height*3/4 snooker-ball-diameter (mm 1))))

(define snooker-cue-ball-start-position
  (vector snooker-table-width/2
	  (- snooker-Baulk (* snooker-D-radius 9/10))))

(define snooker-spot-size (mm 8))

	;;; Balls

(define snooker-cue-ball 0)
(define snooker-yellow-ball 1)
(define snooker-green-ball 2)
(define snooker-brown-ball 3)
(define snooker-blue-ball 4)
(define snooker-pink-ball 5)
(define snooker-black-ball 6)

(define snooker-red-ball-1st  7)
(define snooker-red-ball-last 21)

(define (snooker-red-ball? ball-id)
  (and (<= snooker-red-ball-1st ball-id) (<= ball-id snooker-red-ball-last)))

(define (snooker-ball-value ball-id)
  (cond
   ((not ball-id) 1)
   ((eq? snooker-cue-ball ball-id) 0)
   ((snooker-red-ball? ball-id) 1)
   (else (1+ ball-id))))

(define (snooker-ball-name ball-id)
  (cond
   ((snooker-red-ball? ball-id)
    "red ball")
   ((eq? ball-id snooker-cue-ball)
    "cue ball")
   ((eq? ball-id snooker-yellow-ball)
    "yellow ball")
   ((eq? ball-id snooker-green-ball)
    "green ball")
   ((eq? ball-id snooker-brown-ball)
    "brown ball")
   ((eq? ball-id snooker-blue-ball)
    "blue ball")
   ((eq? ball-id snooker-pink-ball)
    "pink ball")
   ((eq? ball-id snooker-black-ball)
    "black ball")
   (else
    (error "Unknown ball"))))

    
(define (snooker-skins-init)
  (ball-set-skin snooker-cue-ball    "cue ball"    'white-ball  'snooker-ball)

  (ball-set-skin snooker-yellow-ball "yellow ball" 'yellow-ball 'snooker-ball)
  (ball-set-skin snooker-green-ball  "green ball"  'green-ball  'snooker-ball)
  (ball-set-skin snooker-brown-ball  "brown ball"  'brown-ball  'snooker-ball)
  (ball-set-skin snooker-blue-ball   "blue ball"   'blue-ball   'snooker-ball)
  (ball-set-skin snooker-pink-ball   "pink ball"   'pink-ball   'snooker-ball)
  (ball-set-skin snooker-black-ball  "black ball"  'black-ball  'snooker-ball)

  (for/from/to snooker-red-ball-1st snooker-red-ball-last
	       (lambda (x)
		 (ball-set-skin x "red ball" 'red-ball 'snooker-ball))))

(define snooker-positions-alist
  (list
   (list snooker-yellow-ball snooker-yellow-ball-position)
   (list snooker-green-ball snooker-green-ball-position)
   (list snooker-brown-ball snooker-Baulk-spot)
   (list snooker-blue-ball  snooker-Center-spot)
   (list snooker-pink-ball  snooker-Pyramid-spot)
   (list snooker-black-ball snooker-Spot)))

(define snooker-positions-alist^2
  (append snooker-positions-alist snooker-positions-alist))

(define (snooker-balls-init)
  (for-each
   (lambda (pos)
     (place-ball/rr (car pos) (cadr pos)))
   snooker-positions-alist)

  (let ((pos	(make-vector 2))
	(rvec	(make-vector 2 (+ snooker-ball-radius (mm 0.01)))))
    (for/from/to snooker-red-ball-1st snooker-red-ball-last
       (lambda (x)
	 (array-map! pos * rvec (ball-in-pyramid (- x 7)))
	 (array-map! pos + pos snooker-top-red-position)
	 (place-ball/rr x pos)))))

(define (snooker-return-ball-to-table ball-id gs)
  (define (find-adepts lst)
    (cond
     ((null? lst)
      (error "This should not happend"))
     ((eq? ball-id (caar lst))
      lst)
     (else
      (find-adepts (cdr lst)))))

  (define (place-on-spot lst)
    (cond
     ((null? lst)
      #f)
     ((place-ball/rr ball-id (cadar lst))
      #t)
     (else
      (place-on-spot (cdr lst)))))

  (define (place-on-vline x y delta)
    (cond
     ((or (< y 0) (< snooker-table-height y))
      #f)
     ((place-ball/rr ball-id (vector x y))
      #t)
     (else
      (place-on-vline x (+ y delta) delta))))
	
  (define (place-cue-ball vec count)
    (random:solid_sphere! vec)
    (let  ((x	(vector-ref vec 0))
	   (y   (- (abs (vector-ref vec 1)))))
      (cond
        ((place-ball/rr ball-id
			(vector (+ snooker-table-width/2
				   (* x snooker-D-radius))
				(+ snooker-Baulk
				   (* y snooker-D-radius))))
	 #t)
	((= count 0)
	 #f)
	(else
	 (place-cue-ball vec (-1+ count))))))
	
  (cond
   ((snooker-red-ball? ball-id)
    (error "This can not happend :)"))
   ((eq? ball-id snooker-cue-ball)
    (snooker-gs-in-D-ok! gs)

    (or (place-ball/rr ball-id snooker-cue-ball-start-position)
	(place-cue-ball (make-vector 2) 50)
	(ball-in-hand ball-in-hand
		      "You have cue ball in your hand (there is no empty place in D)")))
   (else
    (let* ((adepts	(find-adepts snooker-positions-alist^2))
	   (pos		(cadar adepts))
	   (x		(vector-ref pos 0))
	   (y		(vector-ref pos 1)))

      (or
         ;; try all spots first
        (place-on-spot  adepts)
	 ;; than between ball spot and top (or bottom) cushion
	(place-on-vline x y (cm 0.5))
	(place-on-vline x y (cm -0.5))
	 ;; everything failed - place by hand...
	(ball-in-hand ball-id
		      "All rules for spotting colours faild - please place ball by hand"))))))

	;;; game state 
	  ;  on reds/colours/finish colours 
	  ;  points gained in this shot (negative means penalty)
	  ;  nof red balls on table 
	  ;  nof colours on table 
	  ;  list of colours ball on table
	  ;  cue ball can move inside D shape

(define (snooker-gs-new)
  (vector 'on-reds 0 15
	  (list snooker-yellow-ball snooker-green-ball snooker-brown-ball 
		snooker-blue-ball snooker-pink-ball snooker-black-ball)
	  #t))

(define (snooker-gs-copy gs)
  (vector (snooker-gs-state gs)
	  (snooker-gs-points gs)
	  (snooker-gs-reds gs)
	  (list-copy (snooker-gs-clist gs))
	  (snooker-gs-in-D gs)))

(define (snooker-gs-state gs)
  (vector-ref gs 0))

(define (snooker-gs-points gs)
  (vector-ref gs 1))

(define (snooker-gs-penalty-set! gs penalty)
  (vector-set! gs 1 (min (snooker-gs-points gs) (- penalty))))

(define (snooker-gs-points-reset! gs)
  (vector-set! gs 1 0))

(define (snooker-gs-points-add! gs pts)
	; we should not add points when faul commited
  (if (<= 0 (snooker-gs-points gs))
      (vector-set! gs 1 (+ pts (snooker-gs-points gs)))))

(define (snooker-gs-reds gs)
  (vector-ref gs 2))

(define (snooker-gs-reds-dec! gs)
  (vector-set! gs 2 (-1+ (snooker-gs-reds gs))))

(define (snooker-gs-clist gs)
  (vector-ref gs 3))

(define (snooker-gs-clist-set! gs clist)
  (vector-set! gs 3 clist))

(define (snooker-gs-on-reds! gs)
  (vector-set! gs 0 
	       (if (= 0 (snooker-gs-reds gs))
		   'finish-colours
		   'on-reds)))

(define (snooker-gs-on-next! gs)
  (vector-set! gs 0
	       (cond
		((eq? 'on-reds (snooker-gs-state gs))
		 'on-colours)
		((= 0 (snooker-gs-reds gs))
		 'finish-colours)
		(else
		 'on-reds))))

(define (snooker-gs-on-reds? gs)
  (eq? 'on-reds (snooker-gs-state gs)))

(define (snooker-gs-on-colour? gs)
  (eq? 'on-colours (snooker-gs-state gs)))

(define (snooker-gs-in-D gs)
  (vector-ref gs 4))

(define (snooker-gs-in-D-ok! gs)
  (vector-set! gs 4 #t))

(define (snooker-gs-in-D-reset! gs)
  (vector-set! gs 4 #f))

	;;; Rules

(define (snooker-make-faul game-state . args)
  (game-faul)
  (snooker-gs-penalty-set! game-state (apply max 4 args))
  'faul)

(define (snooker-make-point game-state points)
  (game-reward)
  (snooker-gs-points-add! game-state points))

 (define (snooker-make-potted state)
   (if (not (eq? state 'faul))
       'ball-on-potted
       'faul))

(pool-rules "Snooker"
    'number-of-balls 22
    'number-of-players 2

    'help-file "game_snooker.html"

	 ;;; we have deep-copy func and we do are nice to environment
	   ; so we can decalre safe-to-undo
    'safe-to-undo	#t

    'game-state-deep-copy  snooker-gs-copy

    'table-setup 
      (lambda ()
	(table-new table/snooker)
	(cushion-height (+ snooker-ball-radius (mm 2)))
	(snooker-skins-init)

	(mark-line (vector 0 snooker-Baulk) (vector snooker-table-width snooker-Baulk))
	(mark-arc snooker-Baulk-spot snooker-D-radius 8 180 360)

	(for-each 
	 (lambda (x) (mark-point x snooker-spot-size))
	 (list snooker-Spot snooker-Center-spot snooker-Pyramid-spot snooker-Baulk-spot)))

    'new-game
      (lambda ()
	(snooker-balls-init)

	(place-ball/rr snooker-cue-ball snooker-cue-ball-start-position)
	(place-cue snooker-cue-ball #(-90 0))

		 ; init game state
	(game-state-set! (snooker-gs-new)))

    'score
       '((counter	"Points"))

    'queries
       `((ball
           "Ball on"
	   "Please select your `ball-on'"

	   ,(lambda (round player-id game-state)
	      (eq? 'on-colours (snooker-gs-state game-state)))

	   ,(lambda (round player-id game-state ball-id)
	      (cond
	       ((eq? ball-id snooker-cue-ball)
		"Cue ball can not be your `ball-on'")
	       ((snooker-red-ball? ball-id)
		"Red ball can not be your `ball-on'")
	       (else 
		(game-set-ball-on ball-id)
		(message-start (string "You are on " (snooker-ball-name ball-id)))
		#t)))))

    'next-shot
      (lambda (player-id game-state)
	(case (snooker-gs-state game-state)
	  ((on-reds)
	   (message-start "You are on red ball")
	   (game-set-ball-on snooker-red-ball-1st))
	  ((finish-colours)
	   (game-set-ball-on (car (snooker-gs-clist game-state)))
	   (message-start (string "You are on " 
				  (snooker-ball-name (car (snooker-gs-clist game-state))))))
	  (else
	   (game-set-ball-on snooker-cue-ball)))
	(if (snooker-gs-in-D game-state)
	    (ball-can-move snooker-cue-ball snooker-table-inside-D?)
	    (ball-can-move #f))
	(snooker-gs-points-reset! game-state))

    'end-of-shot
      (lambda (last-state game-state balls-out ball-on)
	(DebugDisplayEndOfShot last-state game-state balls-out)
	(let ((rstate  	   (case last-state
			     ((ball-on-potted)
			      (game-add-point 0 (snooker-gs-points game-state))
			      'next-shot)
			     ((faul)
			      (game-add-penalty 0 (- (snooker-gs-points game-state)))
			      'next-player)
			     ((after-ball-ball-collision)
			      'next-player)
			     ((after-cue-ball-collision)
			      (game-add-penalty 0 (max 4
						     (case (snooker-gs-state game-state)
						       ((on-reds) 1)
						       ((on-colours) (snooker-ball-value ball-on))
						       ((finish-colours) 
							(snooker-ball-value 
							 (car (snooker-gs-clist game-state)))))))
			      'next-player)
			     (else
				 ; HUH???
			      (error "This state" last-state "should not happend!!!")))))
	  (cond
	   ((or (eq? rstate 'end)
		(null? (snooker-gs-clist game-state)))
	    (game-higher-score-wins))
	   (else
	    (case rstate
	      ((next-player)
	       (game-next-player)
	       (snooker-gs-on-reds! game-state))
	      ((next-shot)
	       (snooker-gs-on-next! game-state)))

	    (snooker-gs-in-D-reset! game-state)

	    (for-each 
	     (lambda (ball-id)
	       (snooker-return-ball-to-table ball-id game-state))
	     (sort balls-out >)))))

	(message-end))  ; hide mesage showed in query or next-shot function

    'event-mask '(ball-ball-collision 1   ; only first ball-ball collision is needed ... 
		  cue-ball-collision 
		  ball-jump-off 
		  ball-pocketed)

    'event-handler
      (lambda (event-type info1 info2 state game-state ball-on) ; ball, hole
	(DebugDisplayEvent event-type info1 info2 state game-state "Ball on: " ball-on)
	(case event-type

	  ((cue-ball-collision)
	     (cond

		 ; on start	       

	      ((and (not state) (eq? info1 snooker-cue-ball))
	       'after-cue-ball-collision)

		 ; other than cue ball was hit first

	      ((and (not state) (not (eq? info1 snooker-cue-ball)))
	       (snooker-make-faul game-state 
				  (snooker-ball-value ball-on) 
				  (snooker-ball-value info1)))

	      (else ; state != #f (the beginning ...)
	       (snooker-make-faul game-state 
				  (snooker-ball-value ball-on) 
				  (snooker-ball-value info1)))))

	  ((ball-ball-collision)
	   (case state
	     ((faul)	; just pass
	      'faul)
	     ((after-cue-ball-collision)
	      (let* ((not-a-cue-ball  (if (eq? snooker-cue-ball info1)
					  info2
					  info1))
		     (gs-state	   (snooker-gs-state game-state))
		     (real-ball-on  (case gs-state
				      ((on-reds)
				       snooker-red-ball-1st)
				      ((on-colours)
				       ball-on)
				      ((finish-colours)
				       (car (snooker-gs-clist game-state)))))
		     (correct	   (if (eq? gs-state 'on-reds)
					(snooker-red-ball? not-a-cue-ball)
					(eq? real-ball-on not-a-cue-ball))))
		(if correct
		    'after-ball-ball-collision
		    (snooker-make-faul game-state 
				       (snooker-ball-value not-a-cue-ball)
				       (snooker-ball-value real-ball-on)))))

	     (else ; uninteresting ballxball collision
	      state)))

	  ((ball-pocketed)

	   (if (snooker-red-ball? info1)
	       (snooker-gs-reds-dec! game-state))
	   
	   (case (snooker-gs-state game-state)
		 ; we are playing on reds
	     ((on-reds)
	      (cond 
	       ((snooker-red-ball? info1)
		(snooker-make-point game-state 1)
		(snooker-make-potted state))
	       (else
		(game-ball-out info1)
		(snooker-make-faul game-state 
				   (snooker-ball-value info1)))))

		 ; on colours, finish
	     ((on-colours)
	      (if (not (snooker-red-ball? info1))
		  (game-ball-out info1))
	      (cond
	       ((eq? info1 ball-on)
		(snooker-make-point game-state (snooker-ball-value ball-on))
		(snooker-make-potted state))
	       (else
		(snooker-make-faul game-state 
				   (snooker-ball-value info1) 
				   (snooker-ball-value ball-on)))))

		 ; fin
	     ((finish-colours)
	      (let ((ball-on    (car (snooker-gs-clist game-state))))
		(cond
		 ((eq? ball-on info1)
		  (snooker-gs-clist-set! game-state (cdr (snooker-gs-clist game-state)))
		  (snooker-make-point game-state (snooker-ball-value info1))
		  (snooker-make-potted state))
		 (else
		  (game-ball-out info1)
		  (snooker-make-faul game-state 
				     (snooker-ball-value info1) 
				     (snooker-ball-value info2))))))))

	  ((ball-jump-off)
	   (game-ball-out info1)
	   (snooker-make-faul game-state 
			      (snooker-ball-value info1) 
			      (snooker-ball-value ball-on))))))


