; -*- Scheme -*-

; this file was initialy written for Trapas3D modeler
; thus not eliminated tui (Trapas gUI) prefix

(use-modules (gtk gtk)
	     (gtk gdk))

(define (stringify x)
  (cond
   ((string? x) x)
   ((char? x) (string x))
   ((symbol? x) (symbol->string x))
   ((number? x) (number->string x))
   (else 
    (call-with-output-string (lambda (p) (display x p))))))

(define (gtk-adjustment-goto-end adjust)
  (gtk-adjustment-set-value adjust
	(- (gtk-adjustment-upper adjust) (gtk-adjustment-page-size adjust))))

(define (adjustment-page-scroll adjust mul)
  (let ((lower		(gtk-adjustment-lower adjust))
	(upper		(gtk-adjustment-upper adjust))
	(pagesize	(gtk-adjustment-page-size adjust)))
    
  (gtk-adjustment-set-value adjust
    (max lower
	 (min (- upper pagesize)
	      (+ (gtk-adjustment-value adjust)
		 (* mul pagesize)))))))

(define (console-display widget font . args)
  (gtk-text-freeze widget)
  (for-each (lambda (c) 
	      (gtk-text-insert widget font #f #f c -1))
	    args)
  (gtk-text-thaw widget))

(define tui-console-choose-font
  (let ((italica	(gdk-font-load "-*-helvetica-medium-o-normal-*-*-100-*-*-*-*-*-*"))
	(bold		(gdk-font-load "-*-helvetica-bold-r-normal-*-*-120-*-*-*-*-*-*")))
    (lambda (font)
      (case font
	((bold) bold)
	((italica) italica)
	((default) #f)
	(else (gdk-font-load font))))))

(define current-console #f)

(define-public (tui-console-clear)
  (if current-console
      (let ((text (car current-console)))
	(gtk-text-set-point text 0)
	   (gtk-text-forward-delete text (gtk-text-get-length text)))))

(define-public (tui-console-port)
  (if current-console
    (cadr current-console)
    (current-stdout-port)))

(define-public (tui-console-font font)
  (if current-console
      ((caddr current-console) (tui-console-choose-font font))))

(define (tui-console-new)
  (let* ((vbox		(gtk-vbox-new #f 0))
	 (hbox		(gtk-hbox-new #f 0))
	 (combo		(gtk-combo-new))
	 (tentry	(gtk-combo-entry combo))
	 (adjust	(gtk-adjustment-new 10.0 0.0 10.0 0.1 1.0 1.0))
	 (txtout	(gtk-text-new #f adjust))
	 (scroll	(gtk-vscrollbar-new adjust))

	 (stack-saved?	#f)
	 (the-last-stack #f)

	 (font		#f)
	 (my-port	(make-soft-port 
			  (vector
				; display char
			    (lambda (c) (console-display txtout font (string c)))
				; display string
			    (lambda (s) (console-display txtout font s))
				; flush port
			    (lambda ()  #t)
				; get character
			    (lambda ()  #f)
				; close port
			    (lambda ()  #t))
			  "w"))
	 (console-info (list txtout my-port (lambda (x) (set! font x)))))


    (define (my-display	font . args)
      (apply console-display 
	     txtout (tui-console-choose-font font) args))

    (define (console-eval e-string)
      (lazy-catch #t 
		  (lambda () 
		    (start-stack 'console-stack (eval-string e-string)))
		  lazy-dispatch))

     (define (save-stack)
 	(cond (stack-saved?)
 	      ((not (memq 'debug (debug-options-interface)))
 	       (set! the-last-stack #f)
 	       (set! stack-saved? #t))
 	      (else
 	       (set! the-last-stack 
		 (case (stack-id #t)
		   ((console-stack) (make-stack #t lazy-dispatch 1))
		   (else
		    (make-stack #t lazy-dispatch 0))))
 	       (set! stack-saved? #t))))

    (define (lazy-dispatch . args)
      (save-stack) ; lazy-dispatch eval-string)
      (apply throw args))

		; pack all widgets ...

    (gtk-box-pack-start hbox txtout #t #t 0)
    (gtk-box-pack-end hbox scroll #f #t 0)
    (gtk-box-pack-start vbox hbox #t #t 0)

    (gtk-container-border-width vbox 3)
    (gtk-box-pack-end vbox combo #f #f 5)
    (gtk-combo-set-use-arrows-always combo #t)
    (gtk-combo-disable-activate combo)

    (gtk-widget-show-all vbox)

		; signals - EXECUTE

    (gtk-signal-connect tentry "activate"
       (lambda ()
	 (let ((tinput	 (gtk-entry-get-text tentry))
	       (old-port (current-output-port))
	       (clear   #f))

	   (set! current-console console-info)
	   (set! stack-saved? #f)

	   (set-current-output-port my-port)
	   (my-display 'default tinput "\n") 

	   (catch #t
	      (lambda ()
		(let* ((output (console-eval tinput))
		       (label  (gtk-list-item-new-with-label tinput)))

		  (my-display 'italica " => " (stringify output) "\n")
		  (gtk-list-append-item (gtk-combo-list combo) label)
		  (gtk-widget-show label)
		  (gtk-entry-set-text tentry "")))
	      (lambda (key . args)
		(my-display 'default
			    (call-with-output-string 
			     (lambda (port) (display-backtrace the-last-stack port))))
		(my-display 'bold
			    "ERROR: " (stringify key) "\n"
			    (call-with-output-string
			     (lambda (port)
			       (apply display-error the-last-stack port args))))))
	      
	   (set-current-output-port old-port)
	   (gtk-adjustment-goto-end adjust))))

		; signals - COMPLETION

    (gtk-signal-connect tentry "key_press_event"
       (lambda (key)
	 (let ((keyval (gdk-event-keyval key)))
	   (case keyval
	     ((65289 65366 65365)
	      (case keyval
		((65289)
		 (let* ((str		(gtk-entry-get-text tentry))
			(position	(gtk-editable-get-position tentry))
			(completion (tui-complete str position)))

		   (cond
		    ((null? completion)
		     (gdk-beep))
		    ((string? completion)
		     (gtk-entry-set-position tentry
		       (gtk-editable-insert-text tentry completion position)))
		    (else
		     (for-each (lambda (x) 
				 (console-display txtout #f x)
				 (console-display txtout #f "\n"))
			       completion))))
		 (gtk-adjustment-goto-end adjust))
		((65366) (adjustment-page-scroll adjust 1))
		((65365) (adjustment-page-scroll adjust -1)))
	      (gtk-signal-emit-stop-by-name tentry "key_press_event")
	      #t)
	     (else
	      (tui-complete-clear))))))
	
		; say hallo to user ... :)
   (my-display 'bold "Tulpas")
    (my-display 'italica " - Copyright (c) 2000 Tuhyk Labs\n\n")
    (my-display 'default "Welcome to interactive console...\n\n")

    (set! current-console console-info)
    (cons vbox tentry)))


(define simple-debug-console
  (let ((window #f))
    (lambda ()
      (cond
       (window
	(gtk-widget-show window))
       (else
	(let ((console  (tui-console-new)))
	  (set! window (gtk-window-new 'toplevel))
	  (gtk-window-set-title window "Tulpas buggy console")

	  (gtk-widget-set-usize window 400 300)

	  (gtk-container-border-width window 5)
	  (gtk-container-add window (car console))

	  (gtk-signal-connect window "destroy" 
			      (lambda ()
				(gtk-widget-hide window)))

	  (gtk-widget-show-all window)
	  (gtk-window-set-focus window (cdr console))))))))
