; -*- Scheme -*-

	;;; Balls

(define (ball-random-rotation)
  (vector (- (* (random:uniform) 180) 90)
	  (* (random:uniform) 360)))

(define (place-ball/rr ball-id ball-position)
  (place-ball ball-id ball-position (ball-random-rotation)))

(define (ball-respot-area ball-id lo-x lo-y hi-x hi-y)
  (let ((x-len  (- hi-x lo-x))
	(y-len  (- hi-y lo-y))
	(rot    (ball-random-rotation))
	(v	(vector (* 1/2 (+ lo-x hi-x))
			(* 1/2 (+ lo-y hi-y)))))

    (or (place-ball ball-id v rot)
	(let loop ((iter	100))
	  (vector-set! v 0 (+ lo-x (* x-len (random:uniform))))
	  (vector-set! v 1 (+ lo-y (* y-len (random:uniform))))
	  
	  (unless (place-ball ball-id v rot)
	    (cond
	     ((< 0 iter) (loop (-1+ iter)))
	     (else
	      (ball-in-hand ball-id
			    "Unable to respot ball in area - please do it yourself :)"))))))))

(define (ball-respot ball-id pos lower upper)
  (let ((fpos	 (vector (vector-ref pos 0) #f))
	(start-y (vector-ref pos 1)))

    (call/cc 
     (lambda (break)
       (for/from/to/step start-y upper (cm 0.5)
	  (lambda (x)
	    (vector-set! fpos 1 x)
	    (and (place-ball/rr ball-id fpos)
		 (break #t))))

       (for/from/to/step start-y lower (cm -0.5)
	  (lambda (x)
	    (vector-set! fpos 1 x)
	    (and (place-ball/rr ball-id fpos)
		 (break #t))))
       #f))))

	;;; various standard ball positions

; Pyramid:
;  
;	 10  11  12  13  14
; 	   6   7   8   9
;  ^	     3   4   5
;  |y	       1   2   
;  |	         0
;  |  x
;   ----> 
; ;
; returns:
;  #(dbl dbl) - coordinates n-th ball in pyramid
;   assuming:
;	ball has radius 1)
;	first (zero) ball is on #(0 0)
;	   

(define ball-in-pyramid
  (letrec ((pyramid (lambda (x row balls-in-row)
		   (cond 
		    ((< x balls-in-row)
		     (vector (- (2* x) balls-in-row -1)
			     (* 2 row (cos PI/6))))
		    (else
		     (pyramid (- x balls-in-row) (1+ row) (1+ balls-in-row)))))))
    (lambda (pid)
      (pyramid pid 0 1))))
		     

; Diamond
;
;		8
;             7   5
;	    6   4   2 
;       v-    3   1    -u
;	|\	0      /|
;         \    / \    /
;               60'

; returns:
;  #(dbl dbl) - coordinates n-th ball in diamond 
;   assuming:
;	ball has radius 1)
;	first (zero) ball is on #(0 0)


(use-modules (ice-9 format))

(define (ball-in-diamond n dsize)
  (let* ((u	(modulo n dsize))
	 (v	(floor (/ n dsize)))
	 (s60	(sin PI/3))
	 (c60   (cos PI/3))
	 (res (vector
	       (* 2 c60 (- u v))
	       (* 2 s60 (+ v u)))))
    res))
