; -*- Scheme -*-

	;;; some constants...

(define table-default-spot-size (mm 8))

	;;; table vector:
	  ;  0 table-id - unused ...
	  ;  1 table name
	  ;  2 width
	  ;  3 height
	  ;  4 width/2
	  ;  5 height/2
	  ;  6 head (height * 1/4)
	  ;  7 foot (height * 3/4)
	  ;  8 center-spot
	  ;  9 head-spot
	  ; 10 foot-spot
	  ; 11 new table 
	  ;      (lambda (mark-list)...
	  ; 12   (lambda (x y)  ... - in-the-kitchet?
	  ; 13   ext-freature-alist
	  ; 14   mark-alist

(define (table-name t)     (vector-ref t 1))
(define (table-width t)    (vector-ref t 2))
(define (table-height t)   (vector-ref t 3))
(define (table-width/2 t)  (vector-ref t 4))
(define (table-height/2 t) (vector-ref t 5))
(define (table-head t)     (vector-ref t 6))
(define (table-foot t)     (vector-ref t 7))

(define (table-center-spot t) (vector-ref t  8))
(define (table-head-spot t)   (vector-ref t  9))
(define (table-foot-spot t)   (vector-ref t 10))

(define (table-new t . mark-list)
  ((vector-ref t 11) mark-list))

(define (table-in-the-kitchen-pred t)
  (vector-ref t 12))

(define (table-spot t spot)
  (let ((s (assoc spot (vector-ref t 14))))
    (if (and s (eq? (cadr s) 'spot))
	(caddr s)
	(error "Table %s does not have spot %s" (table-name t) spot))))

	;;; table-gs - table in game state (presumption - index 0)

(define (gs-table-name t)     (vector-ref (vector-ref t 0) 1))
(define (gs-table-width t)    (vector-ref (vector-ref t 0) 2))
(define (gs-table-height t)   (vector-ref (vector-ref t 0) 3))
(define (gs-table-width/2 t)  (vector-ref (vector-ref t 0) 4))
(define (gs-table-height/2 t) (vector-ref (vector-ref t 0) 5))
(define (gs-table-head t)     (vector-ref (vector-ref t 0) 6))
(define (gs-table-foot t)     (vector-ref (vector-ref t 0) 7))

(define (gs-table-center-spot t) (vector-ref (vector-ref t 0) 8))
(define (gs-table-head-spot t)   (vector-ref (vector-ref t 0) 9))
(define (gs-table-foot-spot t)   (vector-ref (vector-ref t 0) 10))

(define (gs-table-table-new t . mark-list)
  ((vector-ref (vector-ref t 0) 11) mark-list))

(define (gs-table-in-the-kitchen-pred t)
  (vector-ref (vector-ref t 0) 12))

(define (gs-table-feature t f)
  (let ((e	(assoc f (vector-ref t 0))))
    (if e
	(cdr e)
	(error "Table %s does not have feature %S" (table-name t) f))))

	;;; table constructor

(define (process-mark-list mark-alist mark-list)
  (when (pair? mark-list)
    (let ((mark-desc (assoc-ref mark-alist (car mark-list))))
      (when mark-desc
	(Write mark-desc)
	(case (car mark-desc)
	  ((spot) (mark-point (cadr mark-desc) table-default-spot-size))
	  ((line) (apply mark-line (cdr mark-desc)))
	  ((arc) (apply mark-arc (cdr mark-desc)))
	  (else
	   (error "Uknown mark type: %s" (car mark-desc)))))
      (process-mark-list mark-alist (cdr mark-list)))))

(define table-list '())

(define (table-new-type name width height geometry-constructor . rest)
  (define (param-ref id default)
    (let ((x	(member id rest)))
      (if x (cadr x) default)))

  (let* ((width/2	(/ width 2))
	 (height/2	(/ height 2))
	 (head		(param-ref 'head (* 1/4 height)))
	 (foot		(param-ref 'foot (* 3/4 height)))

	 (center-spot	(param-ref 'center-spot (vector width/2 height/2)))
	 (head-spot	(param-ref 'head-spot (vector width/2 head)))
	 (foot-spot	(param-ref 'foot-spot (vector width/2 foot)))
	
	 (mark-alist	(apply list
			       (list 'center-spot 'spot center-spot)
			       (list 'foot-spot 'spot foot-spot)
			       (list 'head-spot 'spot head-spot)
			       (list 'kitchen  'line (vector 0 head) (vector width head))
			       (param-ref 'mark-alist '())))
	 (desc   (vector
			  ;  0 table magix
		   '*table*
			  ;  1 table name
		    name
			  ;  2 width
		    width
			  ;  3 height
		    height
			  ;  4 width/2
		    width/2
			  ;  5 height/2
		    height/2
			  ;  6 head (height * 1/4)
		    head
			  ;  7 foot (height * 3/4)
		    foot
			  ;  8 center-spot
		    center-spot
			  ;  9 head-spot
		    head-spot
			  ; 10 foot-spot
		    foot-spot
			  ; 11 new table 
			  ;      (lambda (mark-list)...
		    (lambda (mark-list)
		      (geometry-constructor)
		      (process-mark-list mark-alist mark-list))
			  ; 12   (lambda (x y)  ... - in-the-kitchet?
		    (param-ref 'in-the-kitchen?
			       (lambda (x y)
				 (< y head)))
			  ; 13  ext-feature-alist    
		    (param-ref 'ext-feature-alist '())
		    	  ; 14   mark-alist
		    mark-alist)))

    (set! table-list (cons desc table-list))
    desc))


	;;; keypad views generator

(define (table-keypad-views center w h angle-start view-distance)
  (for-each
   (lambda (id alfa beta)
     (view id (+ alfa angle-start) beta view-distance center))
   '(  1   2   3   6   9   8   7   4)
   '(-45   0  45  90 135 180 225 270)
   '( 45  20  45  20  45  20  45  20))

  (view 5 
	(if (< w h) 0 90)
	90
	; default fovy is 30
	(+ 5 (/ (max w h) (* 2 (tan (deg->rad 15)))))
	center))

	;;; Table wizards...

(define (table-wizard/rectangle w h dw dh dd bwdth)
  (begin-table 4 0
     (vertex (vector 0 0))
     (vertex (vector w 0))
     (vertex (vector w h))
     (vertex (vector 0 h)))

  (table-outline 'default-table-outline-skin
		 (vector (- bwdth) (- bwdth))
		 (vector (+ w bwdth) (- bwdth))
		 (vector (+ w bwdth) (+ h bwdth))
		 (vector (- bwdth) (+ h bwdth)))

  (let ((w/dw	  (/ w (+ dw 1)))
	(h/dh	  (/ h (+ dh 1))))

    (for/from/to 1 dw
       (lambda (x)
	 (let  ((xx   (* x w/dw)))
	   (mark-diamond (vector xx (- dd)) 2 2 'default-diamond-skin)
	   (mark-diamond (vector xx (+ h dd)) 2 2 'default-diamond-skin))))

    (for/from/to 1 dh
       (lambda (y)
	 (let  ((yy   (* y h/dh)))
	   (mark-diamond (vector (- dd)   yy) 2 2 'default-diamond-skin)
	   (mark-diamond (vector (+ w dd) yy) 2 2 'default-diamond-skin))))))

	;;; Tomby's super table constructor

	  ; x	- x dimension of table
	  ; y	- y dimension of table
	  ; r1	- radius of four corner pockets
	  ; r2	- radius of two side holes
	  ; ds	- shift of side holes in x direction
	  ; dd	- shift of cushions
	  ; dr	- vzdalenost vstupni hrany do diry u rohu
	  ; ulfa	- sklon vstupni hrany
	  ; dmr	- vzdalenust vstupni hrany do diry v prostredku
	  ; dmulfa - sklon vstupni hrany
	  ; bwdth - border width

(define (table-wizard/pockets x y r1 r2 ds dd dr ulfa dmr dmulfa bwdth)
  (let ((dR (- dr (/ dd (tan (deg->rad ulfa)))))
	(y/2 (/ y 2))
	(x-dd  (- x dd))
	(y-dd  (- y dd))
	(stepx (/ (- x (* 2 dd)) 4))
	(stepy (/ (- y (* 2 dd)) 8))
	(-r1   (- r1))
	(dmR   (- dmr (/ dd (tan (deg->rad dmulfa))))))
	
  (begin-table 24 6
    (vertex (vector dR 0))
    (vertex (vector dr dd))
    (vertex (vector (- x dr) dd))
    (vertex (vector (- x dR) 0))
    (vertex (vector x dR))
    (vertex (vector x-dd dr))
    (vertex (vector x-dd (- y/2 dmr)))
    (vertex (vector x (- y/2 dmR)))
    (vertex (vector x (+ y/2 dmR)))
    (vertex (vector x-dd (+ y/2 dmr)))
    (vertex (vector x-dd (- y dr)))
    (vertex (vector x (- y dR)))
    (vertex (vector (- x dR) y))
    (vertex (vector (- x dr) y-dd))
    (vertex (vector dr y-dd))
    (vertex (vector dR y))
    (vertex (vector 0 (- y dR)))
    (vertex (vector dd (- y dr)))
    (vertex (vector dd (+ y/2 dmr)))
    (vertex (vector 0 (+ y/2 dmR)))
    (vertex (vector 0 (- y/2 dmR)))
    (vertex (vector dd (- y/2 dmr)))
    (vertex (vector dd dr))
    (vertex (vector 0 dR))

    (pocket (vector 0 0) r1)
    (pocket (vector 0 y) r1)
    (pocket (vector x y) r1)
    (pocket (vector x 0) r1)
    (pocket (vector (- 0 ds) y/2) r2)
    (pocket (vector (+ x ds) y/2) r2)

    (table-outline 'default-table-outline-skin
		   (vector (- bwdth) (- bwdth))
		   (vector (+ x bwdth) (- bwdth))
		   (vector (+ x bwdth) (+ y bwdth))
		   (vector (- bwdth) (+ y bwdth))))

  (for-each (lambda (xx)
	      (mark-diamond xx 2 2 'default-diamond-skin))
	    (list
;	     (vector dd -r1)
	     (vector (+ dd stepx) -r1)
	     (vector (+ dd (* 2 stepx)) -r1)
	     (vector (+ dd (* 3 stepx)) -r1)
;	     (vector (- x dd) -r1)

;	     (vector dd (+ y r1))
	     (vector (+ dd stepx) (+ y r1))
	     (vector (+ dd (* 2 stepx)) (+ y r1))
	     (vector (+ dd (* 3 stepx)) (+ y r1))
;	     (vector (- x dd) (+ y r1))

;	     (vector -r1 dd)
	     (vector -r1 (+ dd (* 1 stepy)))
	     (vector -r1 (+ dd (* 2 stepy)))
	     (vector -r1 (+ dd (* 3 stepy)))
	     (vector -r1 (+ dd (* 5 stepy)))
	     (vector -r1 (+ dd (* 6 stepy)))
	     (vector -r1 (+ dd (* 7 stepy)))
;	     (vector -r1 (+ dd (* 8 stepy)))
;	     (vector -r1 (- y dd))

;	     (vector (+ x r1) dd)
	     (vector (+ x r1) (+ dd (* 1 stepy)))
	     (vector (+ x r1) (+ dd (* 2 stepy)))
	     (vector (+ x r1) (+ dd (* 3 stepy)))
	     (vector (+ x r1) (+ dd (* 5 stepy)))
	     (vector (+ x r1) (+ dd (* 6 stepy)))
	     (vector (+ x r1) (+ dd (* 7 stepy)))
;	     (vector (+ x r1) (+ dd (* 8 stepy)))
;	     (vector (+ x r1) (- y dd))
	     ))))

	;;; tables

(define table/pockets/8ft
  (let ((w	(ft 4))
	(h	(ft 8)))
  (table-new-type "8ft table" w h
    (lambda ()
     (table-wizard/pockets  w h
	 (in 2.75) (in 2.5) 0 (in 1.25) (+ (in 2.75) (cm 3.5))
	 38 (+ (in 2.5) (cm 0.75)) 70 (in 6))
     (table-keypad-views (vector (/ w 2) (/ h 2) 0)
			 w h 90 (ft 8))))))

(define table/pockets/9ft
  (let ((w	(ft 4.5))
	(h	(ft 9)))
    (table-new-type "9ft table"  w h
      (lambda ()
	(table-wizard/pockets  w h
	  (in 2.75) (in 2.5) 0 (in 1.25) (+ (in 2.75) (cm 3.5))
	  38 (+ (in 2.5) (cm 0.75)) 70 (in 6))
	(table-keypad-views (vector (/ w 2) (/ h 2) 0)
			    w h 90 (ft 9))))))

(define table/carom
  (let* ((width	(in 50))
	 (height (in 90))
	 (width/2 (/ width 2))
	 (height/4 (/ height 4)))
		 
  (table-new-type "carom table" width height
     (lambda ()
       (table-wizard/rectangle width height 3 7 (in 2.75) (in 6))
       (table-keypad-views (vector width/2 (/ height 2) 0)
			   width height 90 (in 100)))
     'mark-alist 
       `((carom-spot-1 spot ,(vector (- width/2 (in 6)) height/4))
	 (carom-spot-2 spot ,(vector (+ width/2 (in 6)) height/4))))))
     
(define table/snooker
  (let ((w		(mm 1750))
	(h		(mm 3500)))

   (table-new-type "snooker table" w h 
     (lambda ()
       (table-wizard/pockets w h 
		;;; FIXME !!!
	  (in 2.75) (in 2.75) 0 (in 1) (+ (in 2.75) (cm 4.5))
	  38 (+ (in 2.5) (cm 0.75)) 70 (in 6))
       (table-keypad-views (vector (/ w 2) (/ h 2) 0)
			   w h 90 (m 3.5))))))
