; -*- Scheme -*-

	;;; generally usefull functions and macros

(define PI 3.14159265358979323)
(define PI/2 (/ PI 2))
(define PI/3 (/ PI 3))
(define PI/4 (/ PI 4))
(define PI/6 (/ PI 6))

		;;; translation between various units

	; gram and kilogram

(define (g x)
  x)

(define (kg x)
  (* 1000. x))

	; ounce
(define (oz x)
  (* 28.349523125 x))

	; mm, cm & m

(define (mm x)
  (/ x 10.0))


(define (cm x)
  x)

(define (m x)
  (* x 100))

	; inch & feet

(define (in x) 
  (* 2.54 x))
  
(define (ft x)
  (* 30.48 x))

	; angles

(define (deg->rad deg)
  (* (/ deg 180) PI))

(define (rad->deg rad)
  (* (/ rad PI) 180))

	; colours

(define (colour r g b)
  (vector r g b))

(define (colour255 r g b)
  (vector (/ r 255.) (/ g 255.) (/ b 255.)))

		;;; various usefull macros

(define-macro (begin-table vertices holes . body)
  `(begin
     (table-start ,vertices ,holes)
     ,@body
     (table-end)))

(define-macro (begin-state message . body)
  `(begin
     (state-start ,message)
     ,@body
     (state-end)))

		;;; Emacs Lisp here, emacs lisp there...

(define-macro (unless cond . body)
  `(or ,cond (begin ,@body)))

(define-macro (when cond . body)
  `(and ,cond (begin ,@body)))


		;;; small (but usefull ) functions
		  ; (what a lazy programmer :)

(define (2* x)
  (* 2 x))

(define call/cc call-with-current-continuation)

(define-macro (try . body)
  `(catch #t
     (lambda ()
       ,@body
       #t)
     (lambda args
       #f)))

		;;; for-like functions for Pascal/C lovers...

(define (for/from/to from to f)
  (do ((counter	from (1+ counter)))
      ((> counter to))
    (f counter)))

(define (for/from/to/step from to step f)
  (do ((counter	from (+ counter step)))
      ((> counter to))
    (f counter)))

(define (make-interval from to . step)
  (let ((inc (if (null? step) 1 (car step)))
	(cmp (if (< from to) <= >=)))
    (let loop ((cur	from))
      (if (cmp cur to)
	  (cons cur (loop (+ inc cur)))
	  '()))))

		;;; List functions

(define (list-remove-member lst x)
  (cond
   ((null? lst) '())
   ((eq? x (car lst))
    (list-remove-member (cdr lst) x))
   (else
    (cons (car lst) (list-remove-member (cdr lst) x)))))

		;;; Debug utilities

(define (Write . args)
  (for-each write args)
  (newline))

(define (Display . args)
  (for-each display args)
  (newline))

(define DebugDisplay 
  (if (not tulpas-developers-mode)
      (lambda x #f)
      (lambda args
	(display "-*- ")
	(map display args)
	(display " -*-\n"))))

(define DebugDisplayEndOfShot
  (if (not tulpas-developers-mode)
      (lambda x #f)
      (lambda (last-state game-state ball-out-list)
	(display "-*- End of shot -*-\n")
	(display "\n  last-state:\t") (write last-state)
	(display "\n  game-state:\t") (write game-state)
	(display "\n  ball-out-list:\t") (write ball-out-list)
	(display "\n-*-*-\n"))))

(define DebugDisplayEvent 
  (if (not tulpas-developers-mode)
      (lambda x #f)
      (lambda (event-type info1 info2 state game-state . rest)
	 (letrec ((i1	(case event-type
			   ((ball-ball-collision)
			    "  Ball (1):\t")
			   ((cue-ball-collision 
			     ball-ball-collision 
			     ball-pocketed 
			     ball-jump-off 
			     ball-mantinel-collision)
			    "  Ball:\t")
			   (else
			    "  Info (1):\t")))
		  (i2	(case event-type
			   ((ball-ball-collision)
			    "\n  Ball (2):\t")
			   ((ball-pocketed)
			    "\n  Pocket:\t")
			   ((ball-mantinel-collision)
			    "\n  Mantinel id:\t")
			   (else
			    "\n  Info (2):\t")))
		  (disp-rest (lambda (args)
			       (cond
				((not (null? args))
				 (display "\n  ") (display (car args)) (display ":\t") (write (cadr args))
				 (disp-rest (cddr args)))))))

	   (display "-*-*- Event (") (display event-type) (display ") -*-*-\n")
	  
	   (display i1) (display info1)
	   (display i2) (display info2)
	   (display "\n  State:\t") (write state)
	   (display "\n  GameState:\t") (write game-state)
	   (disp-rest rest)
	   (display "\n-*-*- End of event -*-*-\n")))))

		;;; yet anothor bunch of usefull functions 
  
(define (silly-unsort l)
  (sort l (lambda args (< 100 (random 200)))))

(define (unsort l)
  (silly-unsort (silly-unsort (silly-unsort (silly-unsort l)))))
  
(define (scan-directory-for-games dirname)
  (DebugDisplay "Scanning directory: " dirname)
  (let ((directory  (opendir dirname)))
    (let load-game ((filename	(readdir directory)))
      (cond
       ((eof-object? filename))
       (else
	(and (string? filename) 
	     (> (string-length filename) 4)
	     (string=? ".scm" 
		       (substring filename (- (string-length filename) 4)))
	     (safe-call
	       (lambda ()
		 (DebugDisplay "loading " dirname "/" filename)
		 (load (string dirname "/" filename)))
	       (lambda ()
		 (Display "-!-!-!- Can not load " filename " file -!-!-!-"))))
	(load-game (readdir directory)))))))
